@@ -8,6 +8,7 @@
 # -v version   generates a new changelog section with given version number
 # -d, --fromdirname
 #              Like -v, but takes version from directory name
+# -n, --nmu    Increments version, but for a NMU
 # -p           preserve directory name
 # --closes nnnnn,nnnnn,... Closes bug reports
 # --[no]query  Don't contact the BTS to find the bug report details
@@ -54,22 +55,6 @@ use Getopt::Long;
 use File::Copy;
 use File::Basename;
 use Cwd;
-my $lwp_broken = undef;
-sub have_lwp() {
-    return $lwp_broken if defined $lwp_broken;
-    eval { require LWP::Simple; import LWP::Simple (); };
-    if ($@) {
-	if ($@ =~ /^Can\'t locate LWP\/Simple\.pm/) {
-	    $lwp_broken="the libwww-perl package is not installed";
-	} else {
-	    $lwp_broken="couldn't load LWP::Simple: $@";
-	}
-	# Have to define the functions we're going to be using
-	eval 'sub get ($) {}';
-    }
-    else { $lwp_broken=''; }
-    return $lwp_broken ? 0 : 1;
-}
 
 # Predeclare functions
 sub fatal($);
@@ -97,6 +82,8 @@ Options:
          [Don\'t] try contacting the BTS to get bug titles
   -d, --fromdirname
          Add a new changelog entry with version taken from the directory name
+  -n, --nmu
+         Increment the Debian release number for a non-maintainer upload
   -p, --preserve
          Preserve the directory name
   --no-preserve
@@ -200,7 +187,7 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 # We use bundling so that the short option behaviour is the same as
 # with older debchange versions.
 my ($opt_help, $opt_version);
-my ($opt_i, $opt_a, $opt_v, $opt_d, $opt_D, $opt_u, @closes);
+my ($opt_i, $opt_a, $opt_v, $opt_d, $opt_D, $opt_u, $opt_n, @closes);
 my ($opt_ignore, $opt_level, $opt_regex, $opt_noconf);
 $opt_u = 'low';
 
@@ -215,6 +202,7 @@ GetOptions("help|h" => \$opt_help,
 	   "preserve!" => \$opt_p,
            "D|distribution=s" => \$opt_D,
            "u|urgency=s" => \$opt_u,
+	   "n|nmu" => \$opt_n,
 	   "query!" => \$opt_query,
 	   "closes=s" => \@closes,
 	   "ignore-dirname" => \$opt_ignore,
@@ -264,6 +252,12 @@ until (-f 'debian/changelog') {
     }
 }
 
+# Can't write, so stop now.
+if (! -w 'debian/changelog')
+{
+        die "$progname: debian/changelog is not writable!\n";
+}
+
 #####
 
 # Find the current version number etc.
@@ -410,51 +404,78 @@ else {
 my @closes_text = ();
 my $warnings = 0;
 if (@closes and $opt_query) { # and we have to query the BTS
-    if (! have_lwp()) {
-	warn "$progname warning: $lwp_broken\ncannot query the bug-tracking system\n";
+    if (system('command -v wget >/dev/null 2>&1') >> 8 != 0) {
+	warn "$progname warning: wget not installed, so cannot query the bug-tracking system\n";
 	$opt_query=0;
 	$warnings++;
 	# This will now go and execute the "if (@closes and ! $opt_query)" code
-    } else {
-	foreach my $close (@closes) {
-	    unless ($close =~ /^\d{3,}$/) {
-		warn "$progname warning: Bug number $close is invalid; ignoring\n";
-		$warnings++;
-		next;
-	    }
-	    my $report = get("http://bugs.debian.org/$close");
-	    if (! defined $report) {
-		warn "$progname warning: could not contact BTS about bug \#$close\n";
-		$warnings++;
-		push @closes_text, "Closes: \#$close: \n";
-	    } else {
-		# Parse the HTML.  Yuck.
-		my ($title, $pkg, $srcpkg);
-		if ($report =~ m%<H1>Debian Bug report logs.*?<BR>(.*?)</H1>%) {
-		    $title = unhtmlsanit($1);
-		} else { $title = ''; }
-		if ($report =~ m%Package: <a href="pkgreport.cgi\?pkg=(.*?)">%i) {
-		    $pkg = $1;
-		    if ($report =~ m%Source for $pkg is\s+<a href="pkgreport.cgi\?src=(.*?)">%s) {
-			$srcpkg = $1; }
-		    else { $srcpkg = '<NO SOURCE PACKAGE FOUND>'; }
+    }
+    else
+    {
+	my %bugs;
+	my $lastbug;
 
-		    if ($srcpkg eq $PACKAGE or $pkg eq 'wnpp') {
-			$title =~ s/^($pkg|$srcpkg): //;
-			push @closes_text, "Closes: \#$close: $title\n";
-		    } else {
-			warn "$progname warning: bug \#$close does not appear to belong to this package,\n  disabling closing changelog entry\n";
-			$warnings++;
-			push @closes_text, "Closes?? \#$close: WRONG PACKAGE!! $title\n";
-		    }
-		} else {
-		    warn "$progname warning: couldn't determine package for bug \#$close, so not checking\n";
-		    $warnings++;
-		    push @closes_text, "Closes: \#$close: $title\n";
-		}
+	my $bugs = `wget -q -O - 'http://bugs.debian.org/cgi-bin/pkgreport.cgi?src=$PACKAGE'`;
+	if ($? >> 8 != 0) {
+	    warn "$progname warning: wget failed, so cannot query the bug-tracking system\n";
+	    $opt_query=0;
+	    $warnings++;
+	    # This will now go and execute the "if (@closes and ! $opt_query)" code
+	}
+
+	foreach (split /\n/, $bugs) {
+	    if (m%<a href=\"bugreport.cgi\?bug=([0-9]*).*?>\#\1: (.*?)</a>%) {
+		$bugs{$1} = [$2];
+		$lastbug=$1;
+	    }
+	    elsif (defined $lastbug and
+		   m%<a href=\"pkgreport.cgi\?pkg=([a-z0-9\+\-\.]*)%) {
+		push @{$bugs{$lastbug}}, $1
+		    if exists $bugs{$lastbug};
+		$lastbug = undef;
 	    }
 	}
-    }
+
+	foreach my $close (@closes) {
+	    if (exists $bugs{$close}) {
+		my ($title,$pkg) = @{$bugs{$close}};
+		$title =~ s/^($pkg|$PACKAGE): //;
+		push @closes_text, "Closes: \#$close: $title\n";
+	    }
+	    else { # not our package, or wnpp
+		my $bug = `wget -q -O - 'http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=$close'`;
+		if ($? >> 8 != 0) {
+		    warn "$progname warning: unknown bug \#$close does not belong to $PACKAGE,\n  disabling closing changelog entry\n";
+		    $warnings++;
+		    push @closes_text, "Closes?? \#$close: UNKNOWN BUG IN WRONG PACKAGE!!\n";
+                } else {
+		    my ($bugtitle) = ($bug =~ m%<TITLE>.*?\#$close - (.*?)</TITLE>%);
+		    my ($bugpkg) = ($bug =~ m%<a href=\"pkgreport.cgi\?pkg=([a-z0-9\+\-\.]*)%);
+		    $bugpkg ||= '?';
+		    my ($bugsrcpkg) = ($bug =~ m%<a href=\"pkgreport.cgi\?src=([a-z0-9\+\-\.]*)%);
+		    $bugsrcpkg ||= '?';
+		    if ($bugsrcpkg eq $PACKAGE) {
+			warn "$progname warning: bug \#$close appears to be already archived,\n  disabling closing changelog entry\n";
+			$warnings++;
+			push @closes_text, "Closes?? \#$close: ALREADY ARCHIVED?  $bugtitle!!\n";
+		    }
+		    elsif ($bugpkg eq 'wnpp') {
+			if ($bugtitle =~ /(^(O|RFA|ITA): )/) {
+			    push @closes_text, "New maintainer. (Closes: \#$close: $bugtitle)\n";
+			}
+			elsif  ($bugtitle =~ /(^(RFP|ITP): )/) {
+			    push @closes_text, "Initial release. (Closes: \#$close: $bugtitle)\n";
+			}
+		    }
+		    else {
+			warn "$progname warning: bug \#$close belongs to package $bugpkg (src $bugsrcpkg),\n  not to $PACKAGE: disabling closing changelog entry\n";
+			$warnings++;
+			push @closes_text, "Closes?? \#$close: WRONG PACKAGE!!  $bugtitle\n";
+		    }
+		}
+	    }
+        }
+   }
 }
 
 if (@closes and ! $opt_query) { # and we don't have to query the BTS
@@ -522,7 +543,7 @@ my $tmpchk=1;
 my ($NEW_VERSION, $NEW_SVERSION, $NEW_UVERSION);
 my $line;
 
-if ($opt_i || $opt_v || $opt_d) {
+if ($opt_i || $opt_n || $opt_v || $opt_d) {
     # Check that a given explicit version number is sensible.
     if ($opt_v || $opt_d) {
 	if($opt_v) {
@@ -580,8 +601,16 @@ if ($opt_i || $opt_v || $opt_d) {
 
     if (! $NEW_VERSION) {
 	if ($VERSION =~ /(.*?)([a-yA-Y][a-zA-Z]*|\d*)$/i) {
-	    my $end=$2; $end++;
-	    $NEW_VERSION = "$1$end";
+	    my $end=$2;
+            my $start=$1;
+            # If it's not already an NMU make it so
+            # otherwise we can be safe if we behave like dch -i
+	    if ($opt_n and not $start =~ /\.$/) {
+	    	$end += 0.1;
+            } else {
+                $end++;
+            }
+            $NEW_VERSION = "$start$end";
 	    ($NEW_SVERSION=$NEW_VERSION) =~ s/^\d+://;
 	    ($NEW_UVERSION=$NEW_SVERSION) =~ s/-[^-]*$//;
 	} else {
@@ -593,6 +622,10 @@ if ($opt_i || $opt_v || $opt_d) {
     print O "$PACKAGE ($NEW_VERSION) $distribution; ",
         "urgency=$opt_u\n\n";
 
+    if ($opt_n) {
+        print O "  * Non-maintainer upload.\n";
+        $line = 1;
+    }
     if (@closes_text or $TEXT) {
 	# format it nicely
 	foreach $CHGLINE (@closes_text) {
@@ -604,7 +637,7 @@ if ($opt_i || $opt_v || $opt_d) {
     } else {
 	print O "  * \n";
     }
-    $line = 3;
+    $line += 3;
     print O "\n -- $MAINTAINER <$EMAIL>  $DATE\n\n";
 
     # Copy the old changelog file to the new one
@@ -686,7 +719,7 @@ if ((cwd() =~ m%/\Q$PACKAGE\E-\Q$UVERSION\E$%) && !$opt_p) {
     open PARSED, "dpkg-parsechangelog |"
 	or fatal "Cannot execute dpkg-parsechangelog: $!";
     while (<PARSED>) {
-	if (/^Version:\s(.+?)\s*$/) { $new_version=$1; last; }
+	if (/^Version:\s(.+?)\s*$/) { $new_version=$1; }
     }
 
     close PARSED
