@@ -60,6 +60,8 @@ Usage: $progname [options] [dir ...]
 Options:
     --report, --no-download
                    Only report on newer or absent versions, do not download
+    --debug        Dump the downloaded web pages to stdout for debugging
+                   your watch file. 
     --download     Report on newer and absent versions, and download (default)
     --pasv         Use PASV mode for FTP connections
     --no-pasv      Do not use PASV mode for FTP connections (default)
@@ -320,7 +322,7 @@ for my $dir (@dirs) {
 
 # Now sort the list of directories, so that we process the most recent
 # directories first, as determined by the package version numbers
-@debdirs = Devscripts::Versort::deb_versort(@debdirs);
+@debdirs = Devscripts::Versort::versort(@debdirs);
 
 # Now process the watchfiles in order.  If a directory d has subdirectories
 # d/sd1/debian and d/sd2/debian, which each contain watchfiles corresponding
@@ -594,6 +596,11 @@ EOF
     }
 			
     my $newfile_base=basename($newfile);
+    # Remove HTTP header trash
+    if ($site =~ m%^http://%)
+    {
+        $newfile_base =~ s/\?.*$//;
+    }
     if (! $lastversion or $lastversion eq 'debian') {
 	$lastversion=$pkg_version;
     }
